@@ -1,9 +1,17 @@
+from __future__ import annotations
+
 import sys
+from pathlib import Path
 
 import pytest
 
 IS_PY311 = sys.version_info[:2] >= (3, 11)
 
+if sys.version_info < (3, 8):
+    from typing_extensions import Literal
+else:
+    from typing import Literal
+
 
 @pytest.mark.parametrize("mode", ["normal", "xdist"])
 class TestFixture:
@@ -12,8 +20,8 @@ class TestFixture:
     """
 
     @pytest.fixture
-    def simple_script(self, testdir):
-        testdir.makepyfile(
+    def simple_script(self, pytester: pytest.Pytester) -> None:
+        pytester.makepyfile(
             """
             def test_foo(subtests):
                 for i in range(5):
@@ -22,13 +30,19 @@ class TestFixture:
         """
         )
 
-    def test_simple_terminal_normal(self, simple_script, testdir, mode):
+    def test_simple_terminal_normal(
+        self,
+        simple_script: None,
+        pytester: pytest.Pytester,
+        mode: Literal["normal", "xdist"],
+    ) -> None:
         if mode == "normal":
-            result = testdir.runpytest()
+            result = pytester.runpytest()
             expected_lines = ["collected 1 item"]
         else:
+            assert mode == "xdist"
             pytest.importorskip("xdist")
-            result = testdir.runpytest("-n1")
+            result = pytester.runpytest("-n1")
             expected_lines = ["1 worker [1 item]"]
 
         expected_lines += [
@@ -38,9 +52,14 @@ class TestFixture:
         ]
         result.stdout.fnmatch_lines(expected_lines)
 
-    def test_simple_terminal_verbose(self, simple_script, testdir, mode):
+    def test_simple_terminal_verbose(
+        self,
+        simple_script: None,
+        pytester: pytest.Pytester,
+        mode: Literal["normal", "xdist"],
+    ) -> None:
         if mode == "normal":
-            result = testdir.runpytest("-v")
+            result = pytester.runpytest("-v")
             expected_lines = [
                 "*collected 1 item",
                 "test_simple_terminal_verbose.py::test_foo [[]custom[]] (i=0) SUBPASS *100%*",
@@ -51,8 +70,9 @@ class TestFixture:
                 "test_simple_terminal_verbose.py::test_foo PASSED *100%*",
             ]
         else:
+            assert mode == "xdist"
             pytest.importorskip("xdist")
-            result = testdir.runpytest("-n1", "-v")
+            result = pytester.runpytest("-n1", "-v")
             expected_lines = [
                 "1 worker [1 item]",
                 "*gw0*100%* test_simple_terminal_verbose.py::test_foo*",
@@ -70,8 +90,10 @@ class TestFixture:
         ]
         result.stdout.fnmatch_lines(expected_lines)
 
-    def test_skip(self, testdir, mode):
-        testdir.makepyfile(
+    def test_skip(
+        self, pytester: pytest.Pytester, mode: Literal["normal", "xdist"]
+    ) -> None:
+        pytester.makepyfile(
             """
             import pytest
             def test_foo(subtests):
@@ -82,17 +104,20 @@ class TestFixture:
         """
         )
         if mode == "normal":
-            result = testdir.runpytest()
+            result = pytester.runpytest()
             expected_lines = ["collected 1 item"]
         else:
+            assert mode == "xdist"
             pytest.importorskip("xdist")
-            result = testdir.runpytest("-n1")
+            result = pytester.runpytest("-n1")
             expected_lines = ["1 worker [1 item]"]
         expected_lines += ["* 1 passed, 3 skipped, 2 subtests passed in *"]
         result.stdout.fnmatch_lines(expected_lines)
 
-    def test_xfail(self, testdir, mode):
-        testdir.makepyfile(
+    def test_xfail(
+        self, pytester: pytest.Pytester, mode: Literal["normal", "xdist"]
+    ) -> None:
+        pytester.makepyfile(
             """
             import pytest
             def test_foo(subtests):
@@ -103,15 +128,38 @@ class TestFixture:
         """
         )
         if mode == "normal":
-            result = testdir.runpytest()
+            result = pytester.runpytest()
             expected_lines = ["collected 1 item"]
         else:
+            assert mode == "xdist"
             pytest.importorskip("xdist")
-            result = testdir.runpytest("-n1")
+            result = pytester.runpytest("-n1")
             expected_lines = ["1 worker [1 item]"]
         expected_lines += ["* 1 passed, 3 xfailed, 2 subtests passed in *"]
         result.stdout.fnmatch_lines(expected_lines)
 
+    def test_typing_exported(
+        self, pytester: pytest.Pytester, mode: Literal["normal", "xdist"]
+    ) -> None:
+        pytester.makepyfile(
+            """
+            from pytest_subtests import SubTests
+
+            def test_typing_exported(subtests: SubTests) -> None:
+                assert isinstance(subtests, SubTests)
+            """
+        )
+        if mode == "normal":
+            result = pytester.runpytest()
+            expected_lines = ["collected 1 item"]
+        else:
+            assert mode == "xdist"
+            pytest.importorskip("xdist")
+            result = pytester.runpytest("-n1")
+            expected_lines = ["1 worker [1 item]"]
+        expected_lines += ["* 1 passed *"]
+        result.stdout.fnmatch_lines(expected_lines)
+
 
 class TestSubTest:
     """
@@ -119,8 +167,8 @@ class TestSubTest:
     """
 
     @pytest.fixture
-    def simple_script(self, testdir):
-        return testdir.makepyfile(
+    def simple_script(self, pytester: pytest.Pytester) -> Path:
+        return pytester.makepyfile(
             """
             from unittest import TestCase, main
 
@@ -137,11 +185,15 @@ class TestSubTest:
         )
 
     @pytest.mark.parametrize("runner", ["unittest", "pytest-normal", "pytest-xdist"])
-    def test_simple_terminal_normal(self, simple_script, testdir, runner):
-
+    def test_simple_terminal_normal(
+        self,
+        simple_script: Path,
+        pytester: pytest.Pytester,
+        runner: Literal["unittest", "pytest-normal", "pytest-xdist"],
+    ) -> None:
         suffix = ".test_foo" if IS_PY311 else ""
         if runner == "unittest":
-            result = testdir.run(sys.executable, simple_script)
+            result = pytester.run(sys.executable, simple_script)
             result.stderr.fnmatch_lines(
                 [
                     f"FAIL: test_foo (__main__.T{suffix}) [custom] (i=1)",
@@ -154,11 +206,12 @@ class TestSubTest:
             )
         else:
             if runner == "pytest-normal":
-                result = testdir.runpytest(simple_script)
+                result = pytester.runpytest(simple_script)
                 expected_lines = ["collected 1 item"]
             else:
+                assert runner == "pytest-xdist"
                 pytest.importorskip("xdist")
-                result = testdir.runpytest(simple_script, "-n1")
+                result = pytester.runpytest(simple_script, "-n1")
                 expected_lines = ["1 worker [1 item]"]
             result.stdout.fnmatch_lines(
                 expected_lines
@@ -172,11 +225,15 @@ class TestSubTest:
             )
 
     @pytest.mark.parametrize("runner", ["unittest", "pytest-normal", "pytest-xdist"])
-    def test_simple_terminal_verbose(self, simple_script, testdir, runner):
-
+    def test_simple_terminal_verbose(
+        self,
+        simple_script: Path,
+        pytester: pytest.Pytester,
+        runner: Literal["unittest", "pytest-normal", "pytest-xdist"],
+    ) -> None:
         suffix = ".test_foo" if IS_PY311 else ""
         if runner == "unittest":
-            result = testdir.run(sys.executable, simple_script, "-v")
+            result = pytester.run(sys.executable, simple_script, "-v")
             result.stderr.fnmatch_lines(
                 [
                     f"test_foo (__main__.T{suffix}) ... ",
@@ -190,7 +247,7 @@ class TestSubTest:
             )
         else:
             if runner == "pytest-normal":
-                result = testdir.runpytest(simple_script, "-v")
+                result = pytester.runpytest(simple_script, "-v")
                 expected_lines = [
                     "*collected 1 item",
                     "test_simple_terminal_verbose.py::T::test_foo [[]custom[]] (i=1) SUBFAIL *100%*",
@@ -198,8 +255,9 @@ class TestSubTest:
                     "test_simple_terminal_verbose.py::T::test_foo PASSED *100%*",
                 ]
             else:
+                assert runner == "pytest-xdist"
                 pytest.importorskip("xdist")
-                result = testdir.runpytest(simple_script, "-n1", "-v")
+                result = pytester.runpytest(simple_script, "-n1", "-v")
                 expected_lines = [
                     "1 worker [1 item]",
                     "*gw0*100%* SUBFAIL test_simple_terminal_verbose.py::T::test_foo*",
@@ -218,8 +276,12 @@ class TestSubTest:
             )
 
     @pytest.mark.parametrize("runner", ["unittest", "pytest-normal", "pytest-xdist"])
-    def test_skip(self, testdir, runner):
-        p = testdir.makepyfile(
+    def test_skip(
+        self,
+        pytester: pytest.Pytester,
+        runner: Literal["unittest", "pytest-normal", "pytest-xdist"],
+    ) -> None:
+        p = pytester.makepyfile(
             """
             from unittest import TestCase, main
 
@@ -236,19 +298,23 @@ class TestSubTest:
         """
         )
         if runner == "unittest":
-            result = testdir.runpython(p)
+            result = pytester.runpython(p)
             result.stderr.fnmatch_lines(["Ran 1 test in *", "OK (skipped=3)"])
         else:
             pytest.xfail("Not producing the expected results (#5)")
-            result = testdir.runpytest(p)
+            result = pytester.runpytest(p)  # type:ignore[unreachable]
             result.stdout.fnmatch_lines(
                 ["collected 1 item", "* 3 skipped, 1 passed in *"]
             )
 
     @pytest.mark.parametrize("runner", ["unittest", "pytest-normal", "pytest-xdist"])
     @pytest.mark.xfail(reason="Not producing the expected results (#5)")
-    def test_xfail(self, testdir, runner):
-        p = testdir.makepyfile(
+    def test_xfail(
+        self,
+        pytester: pytest.Pytester,
+        runner: Literal["unittest", "pytest-normal", "pytest-xdist"],
+    ) -> None:
+        p = pytester.makepyfile(
             """
             import pytest
             from unittest import expectedFailure, TestCase, main
@@ -266,18 +332,18 @@ class TestSubTest:
         """
         )
         if runner == "unittest":
-            result = testdir.runpython(p)
+            result = pytester.runpython(p)
             result.stderr.fnmatch_lines(["Ran 1 test in *", "OK (expected failures=3)"])
         else:
-            result = testdir.runpytest(p)
+            result = pytester.runpytest(p)
             result.stdout.fnmatch_lines(
                 ["collected 1 item", "* 3 xfailed, 1 passed in *"]
             )
 
 
 class TestCapture:
-    def create_file(self, testdir):
-        testdir.makepyfile(
+    def create_file(self, pytester: pytest.Pytester) -> None:
+        pytester.makepyfile(
             """
                     import sys
                     def test(subtests):
@@ -299,9 +365,9 @@ class TestCapture:
                 """
         )
 
-    def test_capturing(self, testdir):
-        self.create_file(testdir)
-        result = testdir.runpytest()
+    def test_capturing(self, pytester: pytest.Pytester) -> None:
+        self.create_file(pytester)
+        result = pytester.runpytest()
         result.stdout.fnmatch_lines(
             [
                 "*__ test (i='A') __*",
@@ -321,9 +387,9 @@ class TestCapture:
             ]
         )
 
-    def test_no_capture(self, testdir):
-        self.create_file(testdir)
-        result = testdir.runpytest("-s")
+    def test_no_capture(self, pytester: pytest.Pytester) -> None:
+        self.create_file(pytester)
+        result = pytester.runpytest("-s")
         result.stdout.fnmatch_lines(
             [
                 "start test",
@@ -338,9 +404,11 @@ class TestCapture:
         result.stderr.fnmatch_lines(["hello stderr A", "hello stderr B"])
 
     @pytest.mark.parametrize("fixture", ["capsys", "capfd"])
-    def test_capture_with_fixture(self, testdir, fixture):
-        testdir.makepyfile(
-            r"""
+    def test_capture_with_fixture(
+        self, pytester: pytest.Pytester, fixture: Literal["capsys", "capfd"]
+    ) -> None:
+        pytester.makepyfile(
+            rf"""
             import sys
 
             def test(subtests, {fixture}):
@@ -353,11 +421,9 @@ class TestCapture:
                 out, err = {fixture}.readouterr()
                 assert out == 'start test\nhello stdout A\n'
                 assert err == 'hello stderr A\n'
-        """.format(
-                fixture=fixture
-            )
+            """
         )
-        result = testdir.runpytest()
+        result = pytester.runpytest()
         result.stdout.fnmatch_lines(
             [
                 "*1 passed*",
@@ -366,8 +432,8 @@ class TestCapture:
 
 
 class TestLogging:
-    def create_file(self, testdir):
-        testdir.makepyfile(
+    def create_file(self, pytester: pytest.Pytester) -> None:
+        pytester.makepyfile(
             """
             import logging
 
@@ -385,9 +451,9 @@ class TestLogging:
             """
         )
 
-    def test_capturing(self, testdir):
-        self.create_file(testdir)
-        result = testdir.runpytest("--log-level=INFO")
+    def test_capturing(self, pytester: pytest.Pytester) -> None:
+        self.create_file(pytester)
+        result = pytester.runpytest("--log-level=INFO")
         result.stdout.fnmatch_lines(
             [
                 "*___ test_foo [[]sub2[]] __*",
@@ -395,12 +461,12 @@ class TestLogging:
                 "sub2 stdout",
                 "*-- Captured log call ---*",
                 "INFO     root:test_capturing.py:12 sub2 logging",
-                "*== short test summary info ==*"
+                "*== short test summary info ==*",
             ]
         )
 
-    def test_caplog(self, testdir):
-        testdir.makepyfile(
+    def test_caplog(self, pytester: pytest.Pytester) -> None:
+        pytester.makepyfile(
             """
             import logging
 
@@ -410,21 +476,21 @@ class TestLogging:
 
                 with subtests.test("sub1"):
                     logging.info("inside %s", "subtest1")
-                
+
                 assert len(caplog.records) == 2
                 assert caplog.records[0].getMessage() == "start test"
                 assert caplog.records[1].getMessage() == "inside subtest1"
             """
         )
-        result = testdir.runpytest()
+        result = pytester.runpytest()
         result.stdout.fnmatch_lines(
             [
                 "*1 passed*",
             ]
         )
 
-    def test_no_logging(self, testdir):
-        testdir.makepyfile(
+    def test_no_logging(self, pytester: pytest.Pytester) -> None:
+        pytester.makepyfile(
             """
             import logging
 
@@ -433,7 +499,7 @@ class TestLogging:
 
                 with subtests.test("sub passing"):
                     logging.info("inside %s", "passing log line")
-                
+
                 with subtests.test("sub failing"):
                     logging.info("inside %s", "failing log line")
                     assert False
@@ -441,7 +507,7 @@ class TestLogging:
                 logging.info("end log line")
             """
         )
-        result = testdir.runpytest("-p no:logging")
+        result = pytester.runpytest("-p no:logging")
         result.stdout.fnmatch_lines(
             [
                 "*1 passed*",
@@ -458,34 +524,38 @@ class TestDebugging:
         Fake debugger class implementation that tracks which methods were called on it.
         """
 
-        quitting = False
-        calls = []
+        quitting: bool = False
+        calls: list[str] = []
 
-        def __init__(self, *args, **kwargs):
+        def __init__(self, *_: object, **__: object) -> None:
             self.calls.append("init")
 
-        def reset(self):
+        def reset(self) -> None:
             self.calls.append("reset")
 
-        def interaction(self, *args):
+        def interaction(self, *_: object) -> None:
             self.calls.append("interaction")
 
     @pytest.fixture(autouse=True)
-    def cleanup_calls(self):
+    def cleanup_calls(self) -> None:
         self._FakePdb.calls.clear()
 
-    def test_pdb_fixture(self, testdir, monkeypatch):
-        testdir.makepyfile(
+    def test_pdb_fixture(
+        self, pytester: pytest.Pytester, monkeypatch: pytest.MonkeyPatch
+    ) -> None:
+        pytester.makepyfile(
             """
             def test(subtests):
                 with subtests.test():
                     assert 0
             """
         )
-        self.runpytest_and_check_pdb(testdir, monkeypatch)
+        self.runpytest_and_check_pdb(pytester, monkeypatch)
 
-    def test_pdb_unittest(self, testdir, monkeypatch):
-        testdir.makepyfile(
+    def test_pdb_unittest(
+        self, pytester: pytest.Pytester, monkeypatch: pytest.MonkeyPatch
+    ) -> None:
+        pytester.makepyfile(
             """
             from unittest import TestCase
             class Test(TestCase):
@@ -494,17 +564,46 @@ class TestDebugging:
                         assert 0
             """
         )
-        self.runpytest_and_check_pdb(testdir, monkeypatch)
+        self.runpytest_and_check_pdb(pytester, monkeypatch)
 
-    def runpytest_and_check_pdb(self, testdir, monkeypatch):
+    def runpytest_and_check_pdb(
+        self, pytester: pytest.Pytester, monkeypatch: pytest.MonkeyPatch
+    ) -> None:
         # Install the fake pdb implementation in pytest_subtests so we can reference
         # it in the command line (any module would do).
         import pytest_subtests
 
         monkeypatch.setattr(pytest_subtests, "_CustomPdb", self._FakePdb, raising=False)
-        result = testdir.runpytest("--pdb", "--pdbcls=pytest_subtests:_CustomPdb")
+        result = pytester.runpytest("--pdb", "--pdbcls=pytest_subtests:_CustomPdb")
 
         # Ensure pytest entered in debugging mode when encountering the failing
         # assert.
         result.stdout.fnmatch_lines("*entering PDB*")
         assert self._FakePdb.calls == ["init", "reset", "interaction"]
+
+
+def test_exitfirst(pytester: pytest.Pytester) -> None:
+    """
+    Validate that when passing --exitfirst the test exits after the first failed subtest.
+    """
+    pytester.makepyfile(
+        """
+        def test_foo(subtests):
+            with subtests.test("sub1"):
+                assert False
+
+            with subtests.test("sub2"):
+                assert False
+        """
+    )
+    result = pytester.runpytest("--exitfirst")
+    assert result.parseoutcomes()["failed"] == 2
+    result.stdout.fnmatch_lines(
+        [
+            "*[[]sub1[]] SUBFAIL test_exitfirst.py::test_foo - assert False*",
+            "FAILED test_exitfirst.py::test_foo - assert False",
+            "* stopping after 2 failures*",
+        ],
+        consecutive=True,
+    )
+    result.stdout.no_fnmatch_line("*sub2*")  # sub2 not executed.
